"""The `Message` class and its utility constructors."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, TypeAlias

from ..content import AssistantContentPart, Text, UserContentPart
from ..types import Jsonable

if TYPE_CHECKING:
    from ..clients import ModelId, Provider


@dataclass(kw_only=True)
class SystemMessage:
    """A system message that sets context and instructions for the conversation."""

    role: Literal["system"] = "system"
    """The role of this message. Always "system"."""

    content: Text
    """The content of this `SystemMesssage`."""


@dataclass(kw_only=True)
class UserMessage:
    """A user message containing input from the user."""

    role: Literal["user"] = "user"
    """The role of this message. Always "user"."""

    content: Sequence[UserContentPart]
    """The content of the user message."""

    name: str | None = None
    """A name identifying the creator of this message."""


@dataclass(kw_only=True)
class AssistantMessage:
    """An assistant message containing the model's response."""

    role: Literal["assistant"] = "assistant"
    """The role of this message. Always "assistant"."""

    content: Sequence[AssistantContentPart]
    """The content of the assistant message."""

    name: str | None = None
    """A name identifying the creator of this message."""

    provider: Provider | None
    """The LLM provider that generated this assistant message, if available."""

    model_id: ModelId | None
    """The model identifier of the LLM that generated this assistant message, if available."""

    raw_message: Jsonable | None
    """The provider-specific raw representation of this assistant message, if available.

    If raw_content is truthy, then it may be used for provider-specific behavior when
    resuming an LLM interaction that included this assistant message. For example, we can
    reuse the provider-specific raw encoding rather than re-encoding the message from it's
    Mirascope content representation. This may also take advantage of server-side provider
    context, e.g. identifiers of reasoning context tokens that the provider generated.
    
    If present, the content should be encoded as JSON-serializable data, and in a format
    that matches representation the provider expects representing the Mirascope data.
    This may involve e.g.  converting Pydantic `BaseModel`s into plain dicts via `model_dump`.

    Raw content is not required, as the Mirascope content can also be used to generate
    a valid input to the provider (potentially without taking advantage of provider-specific
    reasoning caches, etc). In that case raw content should be left empty.
    """


Message: TypeAlias = SystemMessage | UserMessage | AssistantMessage
"""A message in an LLM interaction.

Messages have a role (system, user, or assistant) and content that is a sequence
of content parts. The content can include text, images, audio, documents, and
tool interactions.

For most use cases, prefer the convenience functions `system()`, `user()`, and
`assistant()` instead of directly creating `Message` objects.

Example:

    ```python
    from mirascope import llm

    messages = [
        llm.messages.system("You are a helpful assistant."),
        llm.messages.user("Hello, how are you?"),
    ]
    ```
"""

UserContent: TypeAlias = str | UserContentPart | Sequence[str | UserContentPart]
"""Type alias for content that can fit into a `UserMessage`."""

AssistantContent: TypeAlias = (
    str | AssistantContentPart | Sequence[str | AssistantContentPart]
)
"""Type alias for content that can fit into an `AssistantMessage`."""

SystemContent: TypeAlias = str | Text
"""Type alias for content that can fit into a `SystemMessage`."""


def system(content: SystemContent) -> SystemMessage:
    """Creates a system message.

    Args:
        content: The content of the message, which must be a `str` or `Text` content.

    Returns:
        A `SystemMessage`.
    """
    promoted_content = Text(text=content) if isinstance(content, str) else content
    return SystemMessage(content=promoted_content)


def user(
    content: UserContent,
    *,
    name: str | None = None,
) -> UserMessage:
    """Creates a user message.

    Args:
        content: The content of the message, which can be `str` or any `UserContent`,
            or a sequence of such user content pieces.
        name: Optional name to identify a specific user in multi-party conversations.

    Returns:
        A `UserMessage`.
    """
    if isinstance(content, str) or not isinstance(content, Sequence):
        content = [content]
    promoted_content = [
        Text(text=part) if isinstance(part, str) else part for part in content
    ]
    return UserMessage(content=promoted_content, name=name)


def assistant(
    content: AssistantContent,
    *,
    provider: Provider | None,
    model_id: ModelId | None,
    raw_message: Jsonable | None = None,
    name: str | None = None,
) -> AssistantMessage:
    """Creates an assistant message.

    Args:
        content: The content of the message, which can be `str` or any `AssistantContent`,
            or a sequence of assistant content pieces.
        provider: Optional identifier of the provider that produced this message.
        model_id: Optional id of the model that produced this message.
        raw_message: Optional Jsonable object that contains the provider-specific
            "raw" data representation of the content for this assistant message.
        name: Optional name to identify a specific assistant in multi-party conversations.

    Returns:
        An `AssistantMessage`.
    """
    if isinstance(content, str) or not isinstance(content, Sequence):
        content = [content]
    promoted_content = [
        Text(text=part) if isinstance(part, str) else part for part in content
    ]
    return AssistantMessage(
        content=promoted_content,
        provider=provider,
        model_id=model_id,
        raw_message=raw_message,
        name=name,
    )



"""Tests for OpenAICompletionsClient"""

from inline_snapshot import snapshot

from mirascope import llm
from mirascope.llm.clients.openai.responses._utils import encode_request


def test_prepare_message_multiple_assistant_text_parts() -> None:
    """Test preparing an OpenAI request with multiple text parts in an assistant message.

    Included for code coverage.
    """

    messages = [
        llm.messages.user("Hello there"),
        llm.messages.assistant(
            ["General ", "Kenobi"], provider="openai:responses", model_id="gpt-4o"
        ),
    ]
    _, _, kwargs = encode_request(
        model_id="gpt-4o", messages=messages, format=None, tools=None, params={}
    )
    assert kwargs == snapshot(
        {
            "model": "gpt-4o",
            "input": [
                {"role": "user", "content": "Hello there"},
                {"role": "assistant", "content": "General "},
                {"role": "assistant", "content": "Kenobi"},
            ],
        }
    )


def test_prepare_message_multiple_system_messages() -> None:
    """Test preparing an OpenAI request with multiple system messages."""

    messages = [
        llm.messages.user("Hello there"),
        llm.messages.system("Be concise."),
        llm.messages.system("On second thought, be verbose."),
    ]
    _, _, kwargs = encode_request(
        model_id="gpt-4o", messages=messages, format=None, tools=None, params={}
    )
    assert kwargs == snapshot(
        {
            "model": "gpt-4o",
            "input": [
                {"role": "user", "content": "Hello there"},
                {"role": "developer", "content": "Be concise."},
                {"role": "developer", "content": "On second thought, be verbose."},
            ],
        }
    )


def test_context_manager() -> None:
    """Test nested context manager behavior and get_client() integration."""

    global_client = llm.get_client("openai:responses")

    client1 = llm.client("openai:responses", api_key="key1")
    client2 = llm.client("openai:responses", api_key="key2")

    assert llm.get_client("openai:responses") is global_client

    with client1 as ctx1:
        assert ctx1 is client1
        assert llm.get_client("openai:responses") is client1

        with client2 as ctx2:
            assert ctx2 is client2
            assert llm.get_client("openai:responses") is client2

        assert llm.get_client("openai:responses") is client1

    assert llm.get_client("openai:responses") is global_client


def test_client_caching() -> None:
    """Test that client() returns cached instances for identical parameters."""
    client1 = llm.client(
        "openai:responses", api_key="test-key", base_url="https://api.example.com"
    )
    client2 = llm.client(
        "openai:responses", api_key="test-key", base_url="https://api.example.com"
    )
    assert client1 is client2

    client3 = llm.client(
        "openai:responses",
        api_key="different-key",
        base_url="https://api.example.com",
    )
    assert client1 is not client3

    client4 = llm.client(
        "openai:responses",
        api_key="test-key",
        base_url="https://different.example.com",
    )
    assert client1 is not client4

    client5 = llm.client("openai:responses", api_key=None, base_url=None)
    client6 = llm.get_client("openai:responses")
    assert client5 is client6
